//
//	relay6
//
//  ----------------------------------------------------------------------
//	 Copyright 2001-2005
//	Francesco Picasso <s0tp@libero.it>
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2, or (at your option)
//	any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; see the file GPL.txt.  If not, write to
//	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//  ----------------------------------------------------------------------
//


//#define		_WIN32_WINNT	0x0500


#define WIN32_LEAN_AND_MEAN
#include	<windows.h>
#include    <process.h>
#include	<winsock2.h>
#include	<ws2tcpip.h>
#include	<tpipv6.h>
#include    <stdio.h>
#include    <shellapi.h>
#include    "resource.h"


#define     StdExit(m,c)    { Reporter (m,c, 1); __leave; }
#define     PrintUsage()    { usage (); return (-1);  }  


//  here are the DEFAULT options

#define		BUFFER_SIZE			8192
#define		MAX_CONN			10
#define		SERVER_IP			"localhost"
#define     BIND_TUNNEL         "::"

#define     NI_MAXPORT          6       //  the max length of port number
#define     MAX_REPORT_BUF      4096    //  the max report buffer bytes length   

#define     HOST_UNKNOWN        "<unknown>"
#define     BAD_RESOLVE         "<unable to obtain ip>"
#define     ANY_IP              "<any ip>"
#define     ANY_PORT            "<any port>"

#define		WM_ICONIFY			(WM_USER + 1)
#define		WM_ICONMSG			(WM_USER + 2)
#define		WM_HIDEFAKE			(WM_USER + 3)


typedef struct
{
    SOCKET  sock;
    char   *resolved;
    char   *ip;
    char   *bindIp;
    char   *bindPort;
} BNCPARAM;



DWORD WINAPI	    Bnc                 (LPVOID param);
DWORD WINAPI	    BncServer           (LPVOID param);
int				    BncThem             (SOCKET from, SOCKET to);
void			    Reporter            (char *pszAPI, DWORD code, char time);
void			    GracefulCloseSocket (SOCKET s);
void                usage               (void);
BOOL WINAPI         TerminateIt         (DWORD dwCtrlType);
LRESULT CALLBACK    TrayProc            (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp);
DWORD WINAPI        TrayThread          (LPVOID param);

typedef HWND (__stdcall * GETCONSOLEWND_PROC)(void);


BOOL	bTerminate  = FALSE;        //  used to terminate threads
BOOL    bCanIconify = FALSE;        //  used to manage tray events
LONG	NumRespThread;              //  the number of active bnc threads 
char    remoteHost[NI_MAXHOST];     //  remote host string (ip/name)
char    remoteHostBind[NI_MAXHOST]; //  remote bind string (ip/name)
char    remotePortBind[NI_MAXHOST]; //  the default port on exit 
char    remotePort[NI_MAXPORT];     //  remote port string

HWND            hConsoleHWND;       //  the console window handle
HWND            hTrayDialog;        //  the fake dialog window handle
NOTIFYICONDATA  iIconData;          //  systray data
HICON           hIcon;              //  the icon


//  FireWall/Map  
//

#define         MAX_FILE_LEN        64
#define         MAX_MAP_ENTRY       64      //  how many firewall/map entries
#define         MAX_MAP_ATOM_LEN    1024    //  1025 chars

int CreateFireWall  (char* filename);
int IsAllowed       (char* host);

typedef struct  
{
    char    addrIN[NI_MAXHOST];
    char    addrOUT[NI_MAXHOST];
    char    portOUT[NI_MAXPORT];
}
FMAP;

FMAP    FMapData[MAX_MAP_ENTRY];
BOOL    FirewallON;    


static const char *BANNER = "\n relay6 3.0\n\n\tthe tcp bouncer ipv* to ipv*\n\t francesco picasso <s0tp@libero.it>\n\n";


//
//  PARAMS NOTE
//      
//      there are 3 obliged options
//      and 5 optional options in the form /x:yyyy
//
//      obliged
//
//          + server port 
//          + destination host
//          + destination port
//
//      optional
//
//          + bind server on (DEFAULT: 127.0.0.1)       [/b:ip]
//          + maximum connections (DEFAULT: 10)         [/c:num]
//          + bind tunnel on (DEFAULT: not binded)      [/t:ip]
//          + bind tunnel on port (DEFAULT: not binded) [/p:num]
//          + firewall/map file (DEFAULT: none=         [/f:string]
//



//
//  main
//___________________________________________________________________

int main (int argc, char *argv[])       
{	
    char            localHost[NI_MAXHOST];      //  local host string (ip/name)
    char            localPort[NI_MAXPORT];      //  local port string
    char            resolved[NI_MAXHOST];       //  incoming connections' string (resolved)
    char            stringIP[NI_MAXHOST];       //  incoming connections' string (unresolved)
    char            report[MAX_REPORT_BUF];     //  the report error buffer
	WSADATA		    WinsockData;
    int             i;
    int             len;
    LONG            maxCon;
   	SOCKET			server;
	SOCKET			response;
	DWORD			dw;
	ADDRINFO		hint, *AI;
	int				ret;
	fd_set			sset;
	TIMEVAL			timer;
	HANDLE			hRespThread;
	PVOID			temp;
    BNCPARAM        bncParam;
    SOCKADDR_STORAGE    sas;
    GETCONSOLEWND_PROC  GetConsoleWnd;

    //  we init the default params
    //  if the bind strings are empty (xx[0] = 0)
    //  we do not bind
    //
    maxCon = MAX_CONN;
    remoteHostBind[0]   = 0;
    remoteHost[0]       = 0;
    remoteHostBind[0]   = 0;
    remotePortBind[0]   = 0;
    remotePort[0]       = 0;
    CopyMemory (localHost, SERVER_IP, lstrlen (SERVER_IP)+1);

    //  initing the firewall struct
    //
    FirewallON = FALSE;
    for (i=0; i < MAX_MAP_ENTRY; i++)
    {
        FMapData[i].addrIN[0]   = 0;
        FMapData[i].addrOUT[0]  = 0;
        FMapData[i].portOUT[0]  = 0;
    }

    fprintf (stdout, BANNER);

    //  first we parse the command line
    //
    if ((argc <= 1) || (argc > 8))
        PrintUsage ();

    //  we count the obliged options
    //
    dw = 0;
    for (i=1; i<argc; i++)
        if (argv[i][0] != '/')
            dw++;

    //  there are 3 obliged options
    //
    if (dw != 3)
        PrintUsage ();

    //  now we get the options
    //  we use dw to choose which obliged option
    //  we consider: remember that obliged
    //  options' order is port-host-port
    //
    dw = 0;
    for (i=1; i<argc; i++)
    {
        if (argv[i][0] != '/')          //  obliged options
        {
            len = lstrlen (argv[i]);
            switch (dw)
            {
                case 0:
                    if (len >= NI_MAXPORT)
                    {
                        sprintf (report, "PARAM ERROR: server port string too long, must be < %d\n", NI_MAXPORT);
                        Reporter (report, 0, 0);
                        PrintUsage ();
                    }
                    else
                        CopyMemory (localPort, argv[i], ++len);
                break;
                        
                case 1:
                    if (len >= NI_MAXHOST)
                    {
                        sprintf (report, "PARAM ERROR: remote host name too long, must be < %d\n", NI_MAXHOST);
                        Reporter (report, 0, 0);
                        PrintUsage ();
                    }
                    else
                        CopyMemory (remoteHost, argv[i], ++len);
                break;
                
                case 2:
                    if (len >= NI_MAXPORT)
                    {
                        sprintf (report, "PARAM ERROR: remote port string too long, must be < %d\n", NI_MAXPORT);
                        Reporter (report, 0, 0);
                        PrintUsage ();
                    }
                    else
                        CopyMemory (remotePort, argv[i], ++len);
                break;
            }
            dw++;
        }
        else                        //  free options
        {
            len = lstrlen (argv[i]);

            if (len <= 2)           //  first control "/x:" must be 3 at minimum!
            {
                sprintf (report, "PARAM ERROR: %s is wrong, use /x:yyy\n", argv[i]);
                Reporter (report, 0, 0);
                PrintUsage ();
            }

            if (argv[i][2] != ':')  //  second control: the third char must be ':'
            {
                sprintf (report, "PARAM ERROR: %s is wrong, use /x:yyy\n", argv[i]);
                Reporter (report, 0, 0);
                PrintUsage ();
            }

            switch (argv[i][1])
            {
                case 'b':
                    if (len >= (NI_MAXHOST+3))
                    {
                        sprintf (report, "PARAM ERROR: local host bind string too long, must be < %d\n", NI_MAXHOST);
                        Reporter (report, 0, 0);
                        PrintUsage ();
                    }
                    else
                        CopyMemory (localHost, &argv[i][3], (len-2));
                break;

                case 'c':
                    if (len >= (NI_MAXPORT+3))
                    {
                        sprintf (report, "PARAM ERROR: maximum connection string too long, must be < %d\n", NI_MAXPORT);
                        Reporter (report, 0, 0);
                        PrintUsage ();
                    }
                    else
                        maxCon = atoi (&argv[i][3]);                                            
                break;

                case 't':
                    if (len >= (NI_MAXHOST+3))
                    {
                        sprintf (report, "PARAM ERROR: bind tunnel host string too long, must be < %d\n", NI_MAXHOST);
                        Reporter (report, 0, 0);
                        PrintUsage ();
                    }
                    else
                        CopyMemory (remoteHostBind, &argv[i][3], (len-2));
                break;

                case 'p':
                    if (len >= (NI_MAXPORT+3))
                    {
                        sprintf (report, "PARAM ERROR: bind tunnel port string too long, must be < %d\n", NI_MAXPORT);
                        Reporter (report, 0, 0);
                        PrintUsage ();
                    }
                    else
                        CopyMemory (remotePortBind, &argv[i][3], (len-2));
                break;

                case 'f':
                    if (len >= (MAX_FILE_LEN+3))
                    {
                        sprintf (report, "PARAM ERROR: file name too long, must be < %d\n", MAX_FILE_LEN);
                        Reporter (report, 0, 0);
                        PrintUsage ();
                    }
                    else
                    {
                        char filename[MAX_FILE_LEN]; 

                        CopyMemory (filename, &argv[i][3], (len-2));
                        if (!CreateFireWall (filename))
                        {
                            sprintf (report, "Could not create firewall/map!\n");
                            Reporter (report, 0, 0);
                            PrintUsage ();
                        }
                    }
                break;

                default:
                     sprintf (report, "PARAM ERROR: this tag (%c) is not recognized\n", argv[i][1]);   
                     Reporter (report, 0, 0);
                     PrintUsage ();
                break;
            }
        }
    }

    //  now we set the control handler
    //
    if (!SetConsoleCtrlHandler (TerminateIt, TRUE))
    {   
        Reporter ("could not set the console control handler!", GetLastError (), 0);
        return -1;
    }

	bTerminate           = FALSE;
	NumRespThread        = 0;

	SetLastError (0);

    //  at this point we have all the params
    //  and we start with the server thread
    //
    if ( WSAStartup(MAKEWORD(2,2), &WinsockData) != 0)
    {
        Reporter ("WSAStartup", WSAGetLastError (), 0);
        return -1;
    }

	AI = NULL;
	server = INVALID_SOCKET;

    //  we obtain the console window handle
    //
    GetConsoleWnd = (GETCONSOLEWND_PROC) GetProcAddress (GetModuleHandle ("KERNEL32.DLL"), "GetConsoleWindow");
    if (hConsoleHWND = GetConsoleWnd ())    
    {
        hRespThread = (HANDLE) _beginthreadex (0, 0, (LPTHREAD_START_ROUTINE )TrayThread, 0, 0, &dw);
        if ((DWORD)hRespThread != (DWORD)(-1) && hRespThread)
        {   CloseHandle (hRespThread);    }
        else
        {
            Reporter ("_beginthreadex failed -> it's impossible to iconify relay6\n", GetLastError (), 0);
        }
    }
    else
    {
        bCanIconify  = FALSE;
        hConsoleHWND = NULL;
        Reporter ("GetProcAddress failed -> it's impossible to iconify relay6\n", GetLastError (), 0);
    }


    Reporter ("\n----- relay6 started -----\n", 0, 0);
    Reporter (0, 0, 1);
			
	__try
	{
		//	getting the sock address info of the server

		ZeroMemory (&hint, sizeof(hint));
		hint.ai_socktype = SOCK_STREAM;
		hint.ai_flags = AI_PASSIVE;

        if (localHost[0])
        {
            ret = getaddrinfo (localHost, localPort, &hint, &AI);
            if (ret)
            {
                Reporter ("getaddrinfo", ret, 1);
                __leave;
            }
        }
        else            //  if there is a specified ip to bind on
        {
            ret = getaddrinfo (localHost, localPort, &hint, &AI);
            if (ret)
            {
                Reporter ("getaddrinfo", ret, 1);
                __leave;
            }
        }

		//	we do not search the family, because
        //  we rely on the *first* ai, 
        //
		server = socket(AI->ai_family, AI->ai_socktype, AI->ai_protocol);
		if (server == INVALID_SOCKET)
		{
			Reporter ("socket - cannot create server socket", WSAGetLastError (), 1);
			__leave;
		}

		if (bind (server, AI->ai_addr, AI->ai_addrlen) == SOCKET_ERROR)
		{
			Reporter ("bind", WSAGetLastError (), 1);
			__leave;        
        }

		//	I hope 3 will be sufficient
        //
		if (listen (server, 3) == SOCKET_ERROR)
		{
			Reporter ("listen", WSAGetLastError (), 1);
			__leave;
		}

		freeaddrinfo (AI);
		AI = NULL;

		timer.tv_sec  = 1;
		timer.tv_usec = 0;

		do
		{						
			FD_ZERO (&sset);
			FD_SET (server, &sset);
			ret = select (1, &sset, NULL, NULL, &timer);

			if (bTerminate)
				break;

			if (ret == SOCKET_ERROR)
			{
				Reporter ("select", WSAGetLastError (), 1);
				__leave;
			}

			if (FD_ISSET (server, &sset))
			{
                ret = sizeof (sas);
				response = accept (server, (LPSOCKADDR)&sas, &ret);
				if (response == INVALID_SOCKET)
				{
					Reporter ("accept", WSAGetLastError (), 1);
					__leave;
				}

                resolved[0] = 0;
                getnameinfo((LPSOCKADDR)&sas, ret, resolved,sizeof (resolved), NULL, 0, NI_NAMEREQD);

                stringIP[0] = 0;
                dw = sizeof (stringIP);
                WSAAddressToString ((LPSOCKADDR)&sas, ret, NULL, (LPSTR)stringIP, &dw);

   				temp = &NumRespThread;
 				temp = InterlockedCompareExchange (&temp, NULL, NULL);

                if (bTerminate)
                {
                    closesocket (response);
                    break;
                }

				if (*(LONG *)temp >= maxCon)
                {
                    sprintf (report, "connection refused due to maximum connections limit\r\n");
                    send (response, report, lstrlen (report), 0);
					GracefulCloseSocket (response);

                    sprintf (report, "connection from %s [%s] refused due to maximum connections limit [%d/%d]",
                             resolved[0] ? resolved : HOST_UNKNOWN, stringIP[0] ? stringIP : BAD_RESOLVE, *(LONG *)temp, maxCon);

                    Reporter (report, 0, 1);
                }
				else
				{
                    bncParam.sock       = response;
                    bncParam.resolved   = resolved;
                    bncParam.ip         = stringIP;

                    if (FirewallON)
                    {
                        ret = IsAllowed (resolved);
                        if (!ret)
                            ret = IsAllowed (stringIP);

                        if (!ret)
                        {
                            GracefulCloseSocket (response);
                            sprintf (report, "connection from %s [%s] refused due FireWall reject [%d/%d]",
                                    resolved[0] ? resolved : HOST_UNKNOWN, stringIP[0] ? stringIP : BAD_RESOLVE, *(LONG *)temp, maxCon);
    
                            Reporter (report, 0, 1);
                        }
                    }

                    if ((FirewallON && ret) || !FirewallON)
                    {
                        if (FirewallON)
                        {
                            ret--;  //  IsAllowed returns index+1 to FMapData
                            
                            bncParam.bindIp     = FMapData[ret].addrOUT[0] ? FMapData[ret].addrOUT : remoteHostBind;
                            bncParam.bindPort   = FMapData[ret].portOUT[0] ? FMapData[ret].portOUT : remotePortBind;                       
                        }
                        else
                        {
                            bncParam.bindIp     = remoteHostBind;
                            bncParam.bindPort   = remotePortBind;
                        }
                    

                        hRespThread = (HANDLE) _beginthreadex (0, 0, (LPTHREAD_START_ROUTINE )Bnc, (void *)&bncParam, 0, &dw);

                        if ((DWORD)hRespThread != (DWORD)(-1) && hRespThread)
                        {
                            sprintf (report, "%s [%s] is connected",  resolved[0] ? resolved : HOST_UNKNOWN,
                                                                      stringIP[0] ? stringIP : BAD_RESOLVE);
                            Reporter (report, 0, 1);                  
                            CloseHandle (hRespThread);
                        }
                        else
                        {
                            sprintf (report, "connection from %s [%s] failed due thread init error",
                                 resolved[0] ? resolved : HOST_UNKNOWN, stringIP[0] ? stringIP : BAD_RESOLVE);
                            Reporter (report, GetLastError (), 1);      
                            GracefulCloseSocket (response);
                        }
                    }

                    //  this code is useful to report
                    //  a "correct" (from user's point of view)
                    //  actual connections number

                    Sleep (100);
                    temp = &NumRespThread;
                    temp = InterlockedCompareExchange (&temp, NULL, NULL); 
                    sprintf (report, "Actual connections %d\n", (*(LONG *)temp)-1);
                    Reporter (report, 0, 1);
                }
            }
            
		} while (1);
	}
	__finally
	{
		if (server != INVALID_SOCKET)
			GracefulCloseSocket (server);

		if (AI)
			freeaddrinfo (AI);
	}

    //	not a great polling :-/
	//	but it's ok to avoid threads' handle array ;)
    //
    do
    {
        //  SwitchToThread ();
        temp = &NumRespThread;
        temp = InterlockedCompareExchange (&temp, NULL, NULL);
    } while (*(LONG *)temp);

    Reporter (0, 0, 1);
    Reporter ("\n--- relay6 terminated ----\n", 0, 0);

    WSACleanup ();

	return 0;
}



//
//	usage
//		print the usage
//  _________________________________________________________________

void usage (void)
{
    fprintf (stdout, "\nusage: relay6\n\n\tlocalPort remoteHost remotePort /b:xxx /c:xxx /t:xxx /p:xxx /f:xxx\n");
    fprintf (stdout, "\n\tlocalPort\t is the local server's listening port\n");
    fprintf (stdout, "\tremoteHost\t is the remote host (name/ip)\n");
    fprintf (stdout, "\tremotePort\t is the remote host's port\n");
    fprintf (stdout, "\n\t/b:xxx\t\t server binds on xxx (name/ip)    (default %s)\n", SERVER_IP);
    fprintf (stdout, "\t/c:xxx\t\t maximum number of connections    (default %d)\n",MAX_CONN);
    fprintf (stdout, "\t/t:xxx\t\t tunnel binds on ip xxx (name/ip) (default any)\n");
    fprintf (stdout, "\t/p:xxx\t\t tunnel binds on port xxx (num)   (default any)\n");
    fprintf (stdout, "\t/f:xxx\t\t firewall/map file name           (default none)\n");
    fprintf (stdout, "\n");

}



//
//	Bnc
//		the bouncing thread
//  _________________________________________________________________

DWORD WINAPI Bnc (LPVOID param)
{
	SOCKET		from, to;
	int			ret;
	int			pleaseReturn;
	fd_set		sset;
	TIMEVAL		timer;
	ADDRINFO	hint, *AI, *aiB;
    BNCPARAM   *bncParam;
    PVOID       temp;
    char        resolved[NI_MAXHOST];
    char        stringIP[NI_MAXHOST];
    char        bindIP[NI_MAXHOST];
    char        bindPort[NI_MAXPORT];
    char        report[MAX_REPORT_BUF];

	InterlockedIncrement (&NumRespThread);
	bncParam = (BNCPARAM *)param;

    from = bncParam->sock;
    CopyMemory (resolved, bncParam->resolved, lstrlen (bncParam->resolved)+1);
    CopyMemory (stringIP, bncParam->ip, lstrlen (bncParam->ip)+1);
    CopyMemory (bindIP, bncParam->bindIp, lstrlen (bncParam->bindIp)+1);
    CopyMemory (bindPort, bncParam->bindPort, lstrlen (bncParam->bindPort)+1);

	pleaseReturn = 0;
	to = INVALID_SOCKET;
	AI = NULL;
    aiB = NULL;

	__try
	{
		ZeroMemory (&hint, sizeof(hint));
		hint.ai_socktype = SOCK_STREAM;
		ret = getaddrinfo (remoteHost, remotePort, &hint, &AI);
		if (ret)
		{
            sprintf (report, "closing connection for %s [%s] for %s:%s host due to <getaddrinfo> error",
                resolved[0] ? resolved : HOST_UNKNOWN, stringIP[0] ? stringIP : BAD_RESOLVE, remoteHost, remotePort);
			Reporter (report, ret, 1);
			__leave;
		}

		to = socket (AI->ai_family, AI->ai_socktype, AI->ai_protocol);
		if (to == INVALID_SOCKET)
		{
            sprintf (report, "closing connection for %s [%s] due to <socket> error",
                resolved[0] ? resolved : HOST_UNKNOWN, stringIP[0] ? stringIP : BAD_RESOLVE);
			Reporter (report, WSAGetLastError (), 1);			
			__leave;
		}

        //  if we have to bind
        //  on ip/port 
        //
        if (bindIP[0] || bindPort[0])
        {
            ZeroMemory (&hint, sizeof(hint));
            hint.ai_flags = AI_PASSIVE;
            hint.ai_socktype = SOCK_STREAM;            
            ret = getaddrinfo (bindIP, bindPort, &hint, &aiB);
            if (ret)
            {
                sprintf (report, "cannot bind relay6 on this ip [%s] port [%s] <getaddrinfo> error",
                    bindIP[0] ? bindIP : ANY_IP, bindPort[0] ? bindPort : ANY_PORT);
                Reporter (report, ret, 1);
                __leave;
            }

    		if (bind (to, aiB->ai_addr, aiB->ai_addrlen) == SOCKET_ERROR)
	    	{
                sprintf (report, "cannot bind relay6 on this ip [%s] port [%s] <bind> error",
                    bindIP[0] ? bindIP : ANY_IP, bindPort[0] ? bindPort : ANY_PORT);
		    	Reporter (report, WSAGetLastError (), 1);
			    __leave;        
            }

            freeaddrinfo (aiB);
            aiB = NULL;
        }

        if (connect (to, AI->ai_addr, AI->ai_addrlen) == SOCKET_ERROR)
		{	
            sprintf (report, "closing connection for %s [%s] due to <connect> error",
                resolved[0] ? resolved : HOST_UNKNOWN, stringIP[0] ? stringIP : BAD_RESOLVE);
			Reporter (report, WSAGetLastError (), 1);			
			__leave;
		}      

        freeaddrinfo (AI);
        AI = NULL;

		if (bTerminate)
			__leave;

		timer.tv_sec  = 1;
		timer.tv_usec = 0;

		do
		{						
			FD_ZERO (&sset);
			FD_SET (from, &sset);
			FD_SET (to,   &sset);
			ret = select (2, &sset, NULL, NULL, &timer);
			
			if (bTerminate)
				__leave;

			if (ret == SOCKET_ERROR)
			{
                sprintf (report, "closing connection for %s [%s] due to <select> error",
                    resolved[0] ? resolved : HOST_UNKNOWN, stringIP[0] ? stringIP : BAD_RESOLVE);
                Reporter (report, WSAGetLastError (), 1);			
				__leave;
			}

			if (FD_ISSET (from, &sset))
			{
				ret = BncThem (from, to);
								
				if (!ret)
                {
                    sprintf (report, "%s [%s] closed connection... stopping bounce",
                        resolved[0] ? resolved : HOST_UNKNOWN, stringIP[0] ? stringIP : BAD_RESOLVE);
                    Reporter (report, 0, 1);			
					__leave;
                }

				if (ret == SOCKET_ERROR)
				{
                    sprintf (report, "closing connection for %s [%s] due to <send/recv> error",
                        resolved[0] ? resolved : HOST_UNKNOWN, stringIP[0] ? stringIP : BAD_RESOLVE);
                    Reporter (report, WSAGetLastError (), 1);			
					__leave;
				}				
			}

			if (FD_ISSET (to, &sset))
			{
				ret = BncThem (to, from);
								
				if (!ret)
                {
                    Reporter ("remote tunnel endpoint closed connection... stopping bounce", 0, 1);
					__leave;
                }

				if (ret == SOCKET_ERROR)
				{
                    sprintf (report, "closing connection from %s [%s] due to <send/recv> error",
                        resolved[0] ? resolved : HOST_UNKNOWN, stringIP[0] ? stringIP : BAD_RESOLVE);
                    Reporter (report, WSAGetLastError (), 1);			
					__leave;
				}				
			}

		} while (1);
	}

	__finally
	{
		if (AI)
			freeaddrinfo (AI);

		if (aiB)
			freeaddrinfo (aiB);

		if (to != INVALID_SOCKET)
			GracefulCloseSocket (to);

		if (from != INVALID_SOCKET)
			GracefulCloseSocket (from);

        sprintf (report, "the bouncing for %s [%s] is terminated\n",
                 resolved[0] ? resolved : HOST_UNKNOWN, stringIP[0] ? stringIP : BAD_RESOLVE);
        Reporter (report, 0, 1);

		InterlockedDecrement (&NumRespThread);
        
        temp = &NumRespThread;
        temp = InterlockedCompareExchange (&temp, NULL, NULL);
        sprintf (report, "Actual connections %d\n", (*(LONG *)temp)-1);
        Reporter (report, 0, 1);
	}

	return 0;
}



//
//  BncThem
//  this simple routine read from socket <from>
//		and write to socket <to>
//		returns:
//			 SOCKET_ERROR for errors
//			 0            for closed connections
//			 any number	  all is ok
//___________________________________________________________________

int	BncThem (SOCKET from, SOCKET to)
{
	int		ret;
	char	buffer[BUFFER_SIZE];

	ret = recv (from, buffer, BUFFER_SIZE - 1 , 0);
	
	//	if remote host closed the connection
	if (!ret || ret == SOCKET_ERROR)
		 return ret;

	ret = send (to, buffer, ret, 0);

	return ret;
}



//
//	GracefulCloseSocket
//		close gracefully the socket 
//___________________________________________________________________

void GracefulCloseSocket (SOCKET s)
{
	char	b;
	int		ret;

	__try
	{
		if (shutdown (s, SD_BOTH) == SOCKET_ERROR)
			__leave;

		while (ret = recv (s, &b, sizeof b, 0))
				if (!ret || ret == SOCKET_ERROR)
					break;			
	}
	__finally
	{
		closesocket (s);
	}
}



//
//  Terminate
//___________________________________________________________________

BOOL WINAPI TerminateIt (DWORD dwCtrlType)
{
    switch (dwCtrlType)
    {
        case    CTRL_SHUTDOWN_EVENT:
        case    CTRL_LOGOFF_EVENT:
        case    CTRL_CLOSE_EVENT:
        case    CTRL_C_EVENT:

            PostMessage (hTrayDialog, WM_QUIT, 0, 0);
            bTerminate = TRUE;

        break;

        case    CTRL_BREAK_EVENT:
            if (bCanIconify)
                PostMessage (hTrayDialog, WM_ICONIFY, 0, 0);
            else
                Reporter ("Cannot iconify relay6\n", 0, 1);            
        break;

        default:
            return FALSE;
    }

    return TRUE;         
}



//
//	Reporter
//		displays the error number and corresponding message
//___________________________________________________________________

void Reporter (char *msg, DWORD code, char time)
{    
	LPVOID          lpvMessageBuffer = 0;
    SYSTEMTIME      st;
    static WORD     lastHour = 0xffff;

    
    if (time)
    {
        GetLocalTime (&st);

        if (lastHour == 0xffff)
        {
            fprintf (stdout, "[date %d/%d/%d] ", st.wDay, st.wMonth, st.wYear);
            fprintf (stdout, "[%d:%d:%d]\n\n", st.wHour, st.wMinute, st.wSecond);
        }
        else
            if (st.wHour < lastHour)
                fprintf (stdout, "%d/%d/%d\n", st.wDay, st.wMonth, st.wYear);

        if (lastHour != 0xffff)
            fprintf (stdout, "[%d:%d:%d] ", st.wHour, st.wMinute, st.wSecond);

        lastHour = st.wHour;
    }
    
    if (code)
    {
        FormatMessage (	FORMAT_MESSAGE_ALLOCATE_BUFFER|FORMAT_MESSAGE_FROM_SYSTEM,
		                NULL, code,
		                MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
		                (LPTSTR)&lpvMessageBuffer, 0, NULL);

        if (lpvMessageBuffer)
        {
            if (msg)
                fprintf (stdout, ".. ERROR [0x%x] %s - %s\n", code, msg, (char *)lpvMessageBuffer);
            else
                fprintf (stdout, ".. ERROR [0x%x] %s\n", code, (char *)lpvMessageBuffer);
        }
        else
        {
            if (msg)
                fprintf (stdout, ".. ERROR [0x%x] %s - *system error message not avaible*\n", code, msg);
            else
                fprintf (stdout, ".. ERROR [0x%x] *system error message not avaible*\n", code);
        }

        LocalFree (lpvMessageBuffer);
    }
    else
        if (msg)
            fprintf (stdout, "%s\n", msg);
}



//
//  CreateFireWall
//___________________________________________________________________

int CreateFireWall (char* filename)
{
    FILE*   fp;
    int     i,j;
    int     ret;
    char    line[MAX_REPORT_BUF];

    if (!filename)
        return 0;

    fp = fopen (filename, "r");
    if (!fp)
    {
        Reporter ("could not open firewall/map file\n\r", GetLastError (), 0);
        return 0;
    }

    for (i=0, j=0; i <MAX_MAP_ENTRY; i++)
    {
        j++;

        ret = (int) fgets (line, MAX_REPORT_BUF, fp);
        if (!ret)
            break;

        ret = sscanf (line, "%1024s %1024s %5s", FMapData[i].addrIN, FMapData[i].addrOUT, FMapData[i].portOUT);

        if ((i == (MAX_MAP_ENTRY - 1)) && (ret != EOF))
            Reporter ("FireWall/Map ERROR: too many rules!\n", 0, 0);

        if (ret == EOF)
            break;

        if (!ret)
        {
            i--;            
            continue;
        }           

        if (FMapData[i].addrOUT[0] == '0')
            FMapData[i].addrOUT[0] = 0;

        if (FMapData[i].portOUT[0] == '0')
            FMapData[i].portOUT[0] = 0;
    }

    fclose (fp);
    FirewallON = TRUE;

    return 1;
}



//
//  IsAllowed
//
//      return 0 if errors (== not found) otherwise the rules used+1
//___________________________________________________________________

int IsAllowed (char* host)
{
    unsigned int i;

    if (!host)
        return 0;

    i = 0;
    while (FMapData[i].addrIN[0])
    {
        //  the default rule
        //
        if (FMapData[i].addrIN[0] == '0')
            return ++i;

        if (!lstrcmpi (FMapData[i].addrIN, host))
            return ++i;

        i++;
    }
    return 0;
}



//
//	TrayThread
//		the thread responsible to put relay6 in the systray
//  _________________________________________________________________


DWORD WINAPI TrayThread (LPVOID param)
{
    MSG msg;

    __try
    {
        InterlockedIncrement (&NumRespThread);
        hTrayDialog = CreateDialog ((HINSTANCE)GetModuleHandle (0), MAKEINTRESOURCE (IDD_RELAY6TRAY), NULL, TrayProc);

        if (!hTrayDialog)
            __leave;            

        ShowWindow (hTrayDialog, SW_HIDE);
   
        while (1)
        {
            if (PeekMessage (&msg, NULL, 0, 0, PM_NOREMOVE))
            {
                if (!GetMessage (&msg, NULL, 0, 0)) __leave;                    
                TranslateMessage (&msg);
                DispatchMessage (&msg);
            }
            else
                WaitMessage ();
        }	
    }
    __finally
    {
        bCanIconify = FALSE;                       // because the tray thread finished
        InterlockedDecrement (&NumRespThread);
    }
    
    return msg.wParam;
}



//
//	TrayProc
//		the callback window procedure for systray management
//  _________________________________________________________________

LRESULT CALLBACK TrayProc (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{
    switch (msg)
    {
        case WM_INITDIALOG:

            PostMessage (hwnd, WM_HIDEFAKE, 0, 0);       //  we don't show the window
            hTrayDialog = hwnd;

            //  initializating icon data
            //
            iIconData.cbSize    = sizeof (NOTIFYICONDATA);
            iIconData.hWnd      = hwnd;
            iIconData.hIcon     = LoadIcon (GetModuleHandle (0), MAKEINTRESOURCE (IDI_RELAY6));
            lstrcpyn (iIconData.szTip, remoteHost, 63);
            iIconData.uCallbackMessage = WM_ICONMSG;
            iIconData.uFlags = NIF_ICON | NIF_TIP | NIF_MESSAGE;
            iIconData.uID = 0;           

            bCanIconify = TRUE;
            return FALSE;


        case WM_ICONIFY:

            Shell_NotifyIcon (NIM_ADD, &iIconData);
			ShowWindow (hConsoleHWND, 0);					
            return TRUE;


		case WM_ICONMSG:

			if (lp == WM_LBUTTONUP)
			{
                ShowWindow (hConsoleHWND, SW_RESTORE);
                SetForegroundWindow (hConsoleHWND);
                Shell_NotifyIcon (NIM_DELETE, &iIconData);
            }
            return TRUE;

        //  if something goes bad ;)
        //
        case WM_COMMAND:
        {
            switch (LOWORD (wp))
            {
                case IDC_CLOSE:
                    ShowWindow (hwnd, SW_HIDE);
                    return TRUE;
            }
        }          

        case WM_QUIT:

            bCanIconify = FALSE;
            EndDialog (hwnd, 0);
            return FALSE;
    }

    return FALSE;      
}

            
            